import json
import logging

from types import SimpleNamespace

from qolsys.exceptions import UnableToParseEventException
from qolsys.exceptions import UnknownQolsysEventException
from qolsys.exceptions import UnknownQolsysSensorException
from qolsys.partition import QolsysPartition
from qolsys.utils import find_subclass
from qolsys.sensors import QolsysSensor

LOGGER = logging.getLogger(__name__)


class QolsysEvent(object):

    __SUBCLASSES_CACHE = {}

    def __init__(self, request_id: str, raw_event: dict) -> None:
        self._request_id = request_id
        self._raw_event = raw_event

    @property
    def request_id(self):
        return self._request_id

    @property
    def raw(self):
        return self._raw_event

    @property
    def raw_str(self):
        return json.dumps(self.raw)

    @classmethod
    def from_json(cls, data):
        if isinstance(data, str):
            data = json.loads(data)

        event_type = data.get('event')
        if not event_type:
            raise UnknownQolsysEventException(
                f'Event type not found for event {data}'
            )

        klass = find_subclass(cls, event_type, cache=cls.__SUBCLASSES_CACHE)
        if not klass:
            raise UnknownQolsysEventException(
                f"Event type '{event_type}' unsupported for event {data}"
            )

        return klass.from_json(data)


class QolsysEventInfo(QolsysEvent):

    __INFOCLASSES_CACHE = {}

    @classmethod
    def from_json(cls, data):
        event_type = data.get('event')
        if event_type != 'INFO':
            raise UnableToParseEventException(f"Cannot parse event '{event_type}'")

        info_type = data.get('info_type')
        klass = find_subclass(cls, info_type, cache=cls.__INFOCLASSES_CACHE)
        if not klass:
            raise UnknownQolsysEventException(
                f"Event INFO subtype '{info_type}' unsupported "
                f"for event {data}"
            )

        return klass.from_json(data)


class QolsysEventInfoSummary(QolsysEventInfo):

    def __init__(self, partitions: list = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._partitions = partitions

    @property
    def partitions(self):
        return list(self._partitions)

    # @partitions.setter
    # def partitions(self, partitions):
        # self._partitions = partitions

    def __str__(self):
        return (f"<{type(self).__name__} request_id={self.request_id} "
                f"partitions({len(self.partitions)})="
                f"[{', '.join([str(p) for p in self.partitions])}]>")

    @classmethod
    def from_json(cls, data):
        if isinstance(data, str):
            data = json.loads(data)

        info_type = data.get('info_type')
        if info_type != 'SUMMARY':
            raise UnableToParseEventException(
                f"Cannot parse event with info type '{info_type}'")

        return QolsysEventInfoSummary(
            partitions=cls._parse_partitions(data),
            request_id=data.get('requestID'),
            raw_event=data,
        )

    @classmethod
    def _parse_partitions(cls, data):
        partitions = []

        partition_list = data['partition_list']
        for partition_info in partition_list:
            partition = QolsysPartition(
                partition_id=partition_info.get('partition_id'),
                name=partition_info.get('name'),
                status=partition_info.get('status'),
                secure_arm=partition_info.get('secure_arm'),
            )

            zone_list = partition_info['zone_list']
            for sensor_info in zone_list:
                try:
                    partition.add_sensor(QolsysSensor.from_json(sensor_info, partition))
                except UnknownQolsysSensorException:
                    LOGGER.warning(f"sensor of unknown type: {sensor_info}")

            partitions.append(partition)

        return partitions


class QolsysEventInfoSecureArm(QolsysEventInfo):

    def __init__(self, partition_id: int, value: bool, version: int,
                 *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._partition_id = partition_id
        self._value = value
        self._version = version

    @property
    def partition_id(self) -> int:
        return self._partition_id

    @property
    def value(self) -> bool:
        return self._value

    def __str__(self):
        return (f"<{type(self).__name__} request_id={self.request_id} "
                f"partition_id={self.partition_id} value={self.value}>")

    @classmethod
    def from_json(cls, data):
        if isinstance(data, str):
            data = json.loads(data)

        info_type = data.get('info_type')
        if info_type != 'SECURE_ARM':
            raise UnableToParseEventException(
                f"Cannot parse event with info type '{info_type}'")

        return QolsysEventInfoSecureArm(
            partition_id=data.get('partition_id'),
            value=data.get('value'),
            version=data.get('version'),
            request_id=data.get('requestID'),
            raw_event=data,
        )


class QolsysEventZoneEvent(QolsysEvent):

    __ZONEEVENTCLASSES_CACHE = {}

    def __init__(self, version: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._version = version

    @property
    def zone(self):
        return None

    def __str__(self):
        return (f"<{type(self).__name__} "
                f"zone={self.zone} "
                f"version={self._version}>")

    @classmethod
    def from_json(cls, data):
        event_type = data.get('event')
        if event_type != 'ZONE_EVENT':
            raise UnableToParseEventException(f"Cannot parse event '{event_type}'")

        zone_event_type = data.get('zone_event_type')
        if zone_event_type.startswith('ZONE_'):
            zone_event_type = zone_event_type[5:]
        klass = find_subclass(cls, zone_event_type, cache=cls.__ZONEEVENTCLASSES_CACHE)
        if not klass:
            raise UnknownQolsysEventException(
                f"Event ZONE_EVENT subtype '{zone_event_type}' unsupported "
                f"for event {data}"
            )

        return klass.from_json(data)


class QolsysEventZoneEventActive(QolsysEventZoneEvent):

    def __init__(self, zone_id: int, zone_status: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._zone_id = zone_id
        self._zone_status = zone_status

    @property
    def zone(self):
        return SimpleNamespace(
            id=self._zone_id,
            status=self._zone_status,
        )

    @classmethod
    def from_json(cls, data):
        zone_event_type = data.get('zone_event_type')
        if zone_event_type != 'ZONE_ACTIVE':
            raise UnableToParseEventException(
                f"Cannot parse zone event '{zone_event_type}'")

        return QolsysEventZoneEventActive(
            request_id=data.get('requestID'),
            version=data.get('version'),
            zone_id=data.get('zone', {}).get('zone_id'),
            zone_status=data.get('zone', {}).get('status'),
            raw_event=data,
        )


class _QolsysEventZoneEventFullZone(QolsysEventZoneEvent):

    def __init__(self, zone: QolsysSensor, *args, **kwargs) -> None:
        if self.__class__ == _QolsysEventZoneEventFullZone:
            raise RuntimeError('Should not instantiate this class directly')

        super().__init__(*args, **kwargs)

        self._zone = zone

    @property
    def zone(self):
        return self._zone

    @classmethod
    def from_json(cls, data):
        zone_event_type = data.get('zone_event_type')
        if zone_event_type != cls._ZONE_EVENT_TYPE:
            raise UnableToParseEventException(
                f"Cannot parse zone event '{zone_event_type}'")

        zone = data.get('zone')
        try:
            sensor = QolsysSensor.from_json(zone, None)

            return cls(
                request_id=data.get('requestID'),
                version=data.get('version'),
                zone=sensor,
                raw_event=data,
            )
        except UnknownQolsysSensorException:
            LOGGER.warning(f"sensor of unknown type: {zone}")
            raise


class QolsysEventZoneEventUpdate(_QolsysEventZoneEventFullZone):
    _ZONE_EVENT_TYPE = 'ZONE_UPDATE'


class QolsysEventZoneEventAdd(_QolsysEventZoneEventFullZone):
    _ZONE_EVENT_TYPE = 'ZONE_ADD'


class QolsysEventArming(QolsysEvent):

    def __init__(self, partition_id: int, arming_type: str, version: int,
                 delay: int = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._partition_id = partition_id
        self._arming_type = arming_type
        self._version = version
        self._delay = delay

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def arming_type(self):
        return self._arming_type

    @property
    def delay(self):
        return self._delay

    def __str__(self):
        return (f"<{type(self).__name__} "
                f"partition_id={self.partition_id} "
                f"arming_type={self.arming_type} "
                f"delay={self.delay} "
                f"version={self._version}>")

    @classmethod
    def from_json(cls, data):
        event_type = data.get('event')
        if event_type != 'ARMING':
            raise UnableToParseEventException(f"Cannot parse event '{event_type}'")

        return QolsysEventArming(
            request_id=data.get('requestID'),
            version=data.get('version'),
            partition_id=data.get('partition_id'),
            arming_type=data.get('arming_type'),
            delay=data.get('delay'),
            raw_event=data,
        )


class QolsysEventAlarm(QolsysEvent):

    def __init__(self, partition_id: int, alarm_type: str, version: int,
                 *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._partition_id = partition_id
        self._alarm_type = alarm_type or None
        self._version = version

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def alarm_type(self):
        return self._alarm_type

    @property
    def delay(self):
        return self._delay

    def __str__(self):
        return (f"<{type(self).__name__} "
                f"partition_id={self.partition_id} "
                f"alarm_type={self.alarm_type} "
                f"version={self._version}>")

    @classmethod
    def from_json(cls, data):
        event_type = data.get('event')
        if event_type != 'ALARM':
            raise UnableToParseEventException(f"Cannot parse event '{event_type}'")

        return QolsysEventAlarm(
            request_id=data.get('requestID'),
            version=data.get('version'),
            partition_id=data.get('partition_id'),
            alarm_type=data.get('alarm_type'),
            raw_event=data,
        )


class QolsysEventError(QolsysEvent):

    def __init__(self, partition_id: int, error_type: str, description: str,
                 version: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._partition_id = partition_id
        self._error_type = error_type
        self._description = description
        self._version = version

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def error_type(self):
        return self._error_type

    @property
    def description(self):
        return self._description

    def __str__(self):
        return (f"<{type(self).__name__} "
                f"partition_id={self.partition_id} "
                f"error_type={self.error_type} "
                f"description={self.description} "
                f"version={self._version}>")

    @classmethod
    def from_json(cls, data):
        event_type = data.get('event')
        if event_type != 'ERROR':
            raise UnableToParseEventException(f"Cannot parse event '{event_type}'")

        return QolsysEventError(
            request_id=data.get('requestID'),
            version=data.get('version'),
            partition_id=data.get('partition_id'),
            error_type=data.get('error_type'),
            description=data.get('description'),
            raw_event=data,
        )



import asyncio
import json
import logging
import ssl

from qolsys.actions import QolsysAction
from qolsys.actions import QolsysActionInfo
from qolsys.events import QolsysEvent
from qolsys.exceptions import UnknownQolsysEventException
from qolsys.exceptions import UnknownQolsysSensorException
from qolsys.utils import LoggerCallback


LOGGER = logging.getLogger(__name__)


class QolsysSocket(object):
    def __init__(self, hostname: str, port: int = None, token: str = None,
                 logger=None, callback: callable = None,
                 connected_callback: callable = None,
                 disconnected_callback: callable = None,
                 keep_alive: int = None) -> None:
        self._hostname = hostname
        self._port = port or 12345
        self._token = token or ''

        self._logger = logger or LOGGER
        self._callback = callback or LoggerCallback()
        self._connected_callback = connected_callback or LoggerCallback('Connected callback')
        self._disconnected_callback = disconnected_callback or LoggerCallback('Disconnected callback')
        self._keep_alive = keep_alive or 60 * 4  # 4mn, since the panel generally timeouts at 5mn

        self._writer = None

    def create_tasks(self, event_loop):
        return {
            'listen': event_loop.create_task(self.listen()),
            'keep_alive': event_loop.create_task(self.keep_alive()),
        }

    async def send(self, action: QolsysAction):
        if self._writer is None:
            raise Exception('No writer')

        self._logger.debug(f'Sending: {action.with_token(self._token)}')
        self._writer.write(action.with_token(self._token).encode())
        await self._writer.drain()

    async def keep_alive(self):
        while 'we need to keep the connection alive':
            if self._writer is not None:
                self._logger.debug('Sending keep-alive')
                self._writer.write('\n'.encode())
                await self._writer.drain()
            await asyncio.sleep(self._keep_alive)

    async def listen(self):
        # Replace with https://docs.python.org/3/library/ssl.html#ssl.PROTOCOL_TLS_CLIENT ?
        context = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS_CLIENT)
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE

        server = (self._hostname, self._port)

        self._listen = True
        delay_reconnect = 0
        while self._listen:
            writer = None
            try:
                self._logger.info('Establishing connection to '
                                  f'{server[0]}:{server[1]}')
                reader, writer = await asyncio.open_connection(
                    *server, ssl=context, server_hostname='')
                self._writer = writer

                await self.send(QolsysActionInfo())
                await self._connected_callback()

                delay_reconnect = 0
                while 'there is content to read':
                    line = await reader.readline()
                    if not line:
                        self._logger.info('Connection closed by the panel, exiting to reset the connection')
                        break

                    line = line.decode().rstrip('\n')
                    self._logger.debug(f"Data received (len: {len(line)}): {line}")

                    if line == 'ACK':
                        # This is an ACK to a command we sent, we can ignore
                        self._logger.debug('ACK - ignoring.')
                        continue

                    try:
                        # We try to parse the event to one of our event classes
                        event = QolsysEvent.from_json(line)
                    except json.decoder.JSONDecodeError:
                        self._logger.debug(f'Data is not JSON: {line}')
                        continue
                    except UnknownQolsysEventException:
                        self._logger.debug(f'Unknown Qolsys event: {line}')
                        continue
                    except UnknownQolsysSensorException:
                        self._logger.debug(f'Unknown sensor in Qolsys event: {line}')
                        continue

                    try:
                        await self._callback(event)
                    except:  # noqa: E722
                        self._logger.exception(f'Error calling callback for event: {line}')
            except asyncio.exceptions.CancelledError:
                self._listen = False
                self._logger.info('listening cancelled')
            except:  # noqa: E722
                delay_reconnect = min(delay_reconnect * 2 or 1, 60)
                self._logger.exception('error while listening')
            finally:
                await self._disconnected_callback()

                self._writer = None

                if writer:
                    writer.close()
                    try:
                        await writer.wait_closed()
                    except:  # noqa: E722
                        self._logger.exception(
                            'unable to wait for writer to '
                            'be fully closed; this might not be an issue if '
                            'the connection was closed on the other side')

            if self._listen and delay_reconnect:
                self._logger.info(f'sleeping {delay_reconnect} second(s) before reconnecting')
                await asyncio.sleep(delay_reconnect)



import json
import logging
import time

from qolsys.exceptions import UnableToParseSensorException
from qolsys.exceptions import UnknownQolsysSensorException
from qolsys.observable import QolsysObservable
from qolsys.partition import QolsysPartition
from qolsys.utils import find_subclass


LOGGER = logging.getLogger(__name__)


class QolsysSensor(QolsysObservable):
    NOTIFY_UPDATE_PATTERN = 'update_{attr}'
    NOTIFY_UPDATE_STATUS = 'update_status'
    NOTIFY_UPDATE_ATTRIBUTES = 'update_attributes'

    __SUBCLASSES_CACHE = {}
    _common_keys = [
        'name',
        'status',
        'zone_id',
        'partition_id',
    ]
    ATTRIBUTES = [
        'group',
        'state',
        'zone_type',
        'zone_physical_type',
        'zone_alarm_type',
        'tampered',
    ]

    def __init__(self, sensor_id: str, name: str, group: str, status: str,
                 state: str, zone_id: int, zone_type: int,
                 zone_physical_type: int, zone_alarm_type: int,
                 partition_id: int, partition: QolsysPartition) -> None:
        super().__init__()

        self._id = sensor_id
        self._name = name
        self._group = group
        self._status = status
        self._state = state
        self._zone_id = zone_id
        self._zone_type = zone_type
        self._zone_physical_type = zone_physical_type
        self._zone_alarm_type = zone_alarm_type
        self._partition_id = partition_id
        self._partition = partition

        self._tampered = False
        self._last_open_tampered_at = None
        self._last_closed_tampered_at = None

    @property
    def partition(self) -> QolsysPartition:
        return self._partition

    @partition.setter
    def partition(self, partition: QolsysPartition):
        self._partition = partition

    def update(self, sensor: 'QolsysSensor'):
        if self.id != sensor.id:
            LOGGER.warning(f"Updating sensor '{self.id}' ({self.name}) with "
                           f"sensor '{sensor.id}' (different id)")

        # Because any of the attributes might have changed and we want to
        # be able to notify of all of those changes separately and only if they
        # happened, we have to add a bit of smart in there
        attributes_updated = False
        for attr in ['id'] + self._common_keys + self.ATTRIBUTES:
            local_attr = f'_{attr}'
            prev_value = getattr(self, local_attr)
            new_value = getattr(sensor, attr)
            if prev_value != new_value:
                setattr(self, local_attr, new_value)
                self.notify(change=self.NOTIFY_UPDATE_PATTERN.format(attr=attr),
                            prev_value=prev_value, new_value=new_value)

                if attr in self.ATTRIBUTES:
                    attributes_updated = True

        if attributes_updated:
            self.notify(change=self.NOTIFY_UPDATE_ATTRIBUTES)

    @property
    def id(self):
        return self._id

    @property
    def unique_id(self):
        # Check if this sensor's zone_id is the same as the first sensor's
        # zone_id we have seen with this id. If it is, we return the sensor
        # id directly, if not, we will want to append the zone_id to ensure
        # distinct sensor unique ids
        if self._partition is None:
            raise AttributeError("Partition not set for sensor")

        first_sensor = self._partition.sensor(self._id)
        if first_sensor is None or first_sensor.zone_id != self._zone_id:
            return f"{self._id}_{self._zone_id}"
        return self.id

    @property
    def name(self):
        return self._name

    @property
    def group(self):
        return self._group

    @property
    def status(self):
        return self._status

    @property
    def state(self):
        return self._state

    @property
    def zone_id(self):
        return self._zone_id

    @property
    def zone_type(self):
        return self._zone_type

    @property
    def zone_physical_type(self):
        return self._zone_physical_type

    @property
    def zone_alarm_type(self):
        return self._zone_alarm_type

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def tampered(self):
        return self._tampered

    @property
    def is_open(self):
        return self._status == 'Open'

    @property
    def is_closed(self):
        return self._status == 'Closed'

    @status.setter
    def status(self, value):
        new_value = value.capitalize()
        if new_value not in ['Open', 'Closed']:
            raise AttributeError(f"Invalid value '{value}' for attribute 'status'")

        if self._status != new_value:
            LOGGER.debug(f"Sensor '{self.id}' ({self.name}) status updated to '{new_value}'")
            prev_value = self._status

            self._status = new_value

            self.notify(change=self.NOTIFY_UPDATE_STATUS,
                        prev_value=prev_value, new_value=new_value)

    @tampered.setter
    def tampered(self, value):
        new_value = bool(value)

        if self._tampered != new_value:
            LOGGER.debug(f"Sensor '{self.id}' ({self.name}) tampered updated to '{new_value}'")
            prev_value = self._tampered

            self._tampered = new_value

            self.notify(change=self.NOTIFY_UPDATE_PATTERN.format(attr='tampered'),
                        prev_value=prev_value, new_value=new_value)
            self.notify(change=self.NOTIFY_UPDATE_ATTRIBUTES)

    def _next_status_update_is_status(self):
        # When we are back from a tamper setting, we get two updates
        # subsequently as an open, and then a close, in the same second
        return (self._last_open_tampered_at is not None and
                self._last_closed_tampered_at is not None and
                self._last_closed_tampered_at - self._last_open_tampered_at < 1)

    def open(self):
        if self.is_open and not self._next_status_update_is_status():
            self._last_open_tampered_at = time.time()
            self.tampered = True
        else:
            self.status = 'Open'

    def closed(self):
        if self.tampered:
            self._last_closed_tampered_at = time.time()
            self.tampered = False
        else:
            self.status = 'Closed'

    def __str__(self):
        return (f"<{type(self).__name__} id={self.id} name={self.name} "
                f"group={self.group} status={self.status} "
                f"state={self.state} zone_id={self.zone_id} "
                f"zone_type={self.zone_type} "
                f"zone_physical_type={self.zone_physical_type} "
                f"zone_alarm_type={self.zone_alarm_type} "
                f"partition_id={self.partition_id}>")

    @classmethod
    def from_json(cls, data, partition):
        if isinstance(data, str):
            data = json.loads(data)

        sensor_type = data.get('type')
        if not sensor_type:
            raise UnknownQolsysSensorException(
                f'Sensor type not found for sensor {data}'
            )

        klass = find_subclass(cls, sensor_type, cache=cls.__SUBCLASSES_CACHE,
                              preserve_capitals=True)
        if not klass:
            raise UnknownQolsysSensorException(
                f"Sensor type '{sensor_type}' unsupported for sensor {data}"
            )

        return klass.from_json(data, partition, common=cls.from_json_common_data(data))

    @classmethod
    def from_json_common_data(cls, data):
        common_data = {k: v for k, v in data.items()
                       if k in cls._common_keys or k in cls.ATTRIBUTES}
        common_data['sensor_id'] = data['id']
        return common_data

    @classmethod
    def from_json_subclass(cls, subtype, data, partition, common=None):
        sensor_type = data.get('type')
        if sensor_type != subtype:
            raise UnableToParseSensorException(
                f"Cannot parse sensor '{sensor_type}'")

        if common is None:
            common = cls.from_json_common_data(data)

        return cls(partition=partition, **common)


class _QolsysSensorWithoutUpdates(object):
    pass


class QolsysSensorDoorWindow(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Door_Window', data, partition, common)


class QolsysSensorMotion(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Motion', data, partition, common)


class QolsysSensorPanelMotion(QolsysSensorMotion):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Panel Motion', data, partition, common)


class QolsysSensorGlassBreak(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('GlassBreak', data, partition, common)


class QolsysSensorPanelGlassBreak(QolsysSensorGlassBreak, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Panel Glass Break', data, partition, common)


class QolsysSensorBluetooth(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Bluetooth', data, partition, common)


class QolsysSensorSmokeDetector(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('SmokeDetector', data, partition, common)


class QolsysSensorCODetector(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('CODetector', data, partition, common)


class QolsysSensorWater(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Water', data, partition, common)


class QolsysSensorFreeze(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Freeze', data, partition, common)


class QolsysSensorHeat(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Heat', data, partition, common)


class QolsysSensorTilt(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Tilt', data, partition, common)


class QolsysSensorKeypad(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Keypad', data, partition, common)


class QolsysSensorAuxiliaryPendant(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Auxiliary Pendant', data, partition, common)


class QolsysSensorSiren(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Siren', data, partition, common)


class QolsysSensorKeyFob(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('KeyFob', data, partition, common)


class QolsysSensorTemperature(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Temperature', data, partition, common)


class QolsysSensorTakeoverModule(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('TakeoverModule', data, partition, common)


class QolsysSensorTranslator(QolsysSensor, _QolsysSensorWithoutUpdates):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Translator', data, partition, common)


class QolsysSensorDoorbell(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Doorbell', data, partition, common)


class QolsysSensorShock(QolsysSensor):
    @classmethod
    def from_json(cls, data, partition, common=None):
        return cls.from_json_subclass('Shock', data, partition, common)



import logging

LOGGER = logging.getLogger(__name__)


class QolsysObservable(object):
    def __init__(self):
        self._observers = dict()

    def register(self, observer, callback=None):
        LOGGER.debug(f"Registering {repr(observer)} to {self} updates")
        if callback is None:
            callback = getattr(observer, 'update')
        self._observers[observer] = callback

    def unregister(self, observer):
        LOGGER.debug(f"Unregistering {repr(observer)} from {self} updates")
        del self._observers[observer]

    def notify(self, **payload):
        LOGGER.debug(f"Notifying {self} observers with: {payload}")
        for observer, callback in self._observers.items():
            callback(self, **payload)



import logging

from qolsys.exceptions import QolsysGwConfigIncomplete
from qolsys.exceptions import QolsysGwConfigError
from qolsys.utils import get_mac_from_host


LOGGER = logging.getLogger(__name__)


class QolsysGatewayConfig(object):

    _SENTINEL = object()
    _DEFAULT_CONFIG = {
        'panel_host': _SENTINEL,
        'panel_port': None,
        'panel_mac': None,
        'panel_token': _SENTINEL,
        'panel_user_code': None,
        'panel_unique_id': 'qolsys_panel',
        'panel_device_name': 'Qolsys Panel',
        'arm_away_exit_delay': None,
        'arm_stay_exit_delay': None,
        'arm_away_bypass': None,
        'arm_stay_bypass': None,
        'arm_type_custom_bypass': 'arm_away',

        'mqtt_namespace': 'mqtt',
        'mqtt_retain': True,
        'discovery_topic': 'homeassistant',
        'control_topic': '{discovery_topic}/alarm_control_panel/{panel_unique_id}/set',
        'event_topic': 'qolsys/{panel_unique_id}/event',
        'user_control_token': None,

        'ha_check_user_code': True,
        'ha_user_code': None,
        'code_arm_required': False,
        'code_disarm_required': False,
        'code_trigger_required': False,
        'default_trigger_command': None,
        'default_sensor_device_class': 'safety',
        'enable_static_sensors_by_default': False,
    }

    def __init__(self, args=None, check=True):
        self._override_config = {}

        if args:
            self.load(args)

        if check:
            self.check()

    def load(self, args):
        for k, v in args.items():
            if v is not None and k in self._DEFAULT_CONFIG:
                self._override_config[k] = v

    def check(self):
        errors = 0
        for k in self._DEFAULT_CONFIG.keys():
            if self.get(k) is self._SENTINEL:
                LOGGER.error(f"Missing mandatory configuration key '{k}'")
                errors += 1
        if errors > 0:
            raise QolsysGwConfigIncomplete

        if self.get('panel_user_code') is None:
            if self.get('ha_user_code'):
                raise QolsysGwConfigError("Cannot use 'ha_user_code' if "
                                          "'panel_user_code' is not set")

            for k in ['code_arm_required', 'code_trigger_required']:
                if self.get(k):
                    raise QolsysGwConfigError(
                        f"Cannot use '{k}' if no disarm code is set, as the "
                        "Qolsys Panel does not offer a built-in way to check "
                        "for the code on ARM or TRIGGER actions.")

            # Without a configured disarm code, we cannot have home assistant
            # checking it for us
            self._override_config['ha_check_user_code'] = False

            # Without a configured disarm code, we will use the one provided
            # in home assistant to try and disarm the alarm
            self._override_config['code_disarm_required'] = True

        trig_cmd = self.get('default_trigger_command')
        if trig_cmd:
            trig_cmd = trig_cmd.upper()
            valid_trigger = [
                'TRIGGER',
                'TRIGGER_AUXILIARY',
                'TRIGGER_FIRE',
                'TRIGGER_POLICE',
            ]
            if trig_cmd not in valid_trigger:
                raise QolsysGwConfigError(
                    f"Invalid trigger command '{trig_cmd}'; must be one of "
                    f"{', '.join(valid_trigger)}")

            self._override_config['default_trigger_command'] = trig_cmd

        arm_type = self.get('arm_type_custom_bypass')
        if arm_type:
            arm_type = arm_type.lower()
        valid_arm_type = [
            'arm_stay',
            'arm_away',
        ]
        if arm_type not in valid_arm_type:
            raise QolsysGwConfigError(
                f"Invalid arm type '{arm_type}' for custom bypass; must be "
                f"one of {', '.join(valid_arm_type)}")
        self._override_config['arm_type_custom_bypass'] = arm_type

        if self.get('panel_mac') is None:
            mac = get_mac_from_host(self.get('panel_host'))
            if mac:
                self._override_config['panel_mac'] = mac

        # Apply a template to the control and event topics if the unique id
        # is part of the requested topics
        for k in ('control_topic', 'event_topic'):
            v = self.get(k)
            if v:
                self._override_config[k] = v.format(
                    panel_unique_id=self.get('panel_unique_id'),
                    discovery_topic=self.get('discovery_topic'))

        # Make sure that user codes are stored as strings
        for k in ('panel_user_code', 'ha_user_code'):
            v = self.get(k)
            if v and not isinstance(v, str):
                self._override_config[k] = str(v)

    def get(self, name):
        value = self._override_config.get(name, self._SENTINEL)
        if value is self._SENTINEL:
            value = self._DEFAULT_CONFIG.get(name, self._SENTINEL)
        return value

    def __getattr__(self, name):
        value = self.get(name)
        if value is self._SENTINEL:
            raise AttributeError(f'Parameter `{name}` has not been defined')
        return value



import logging

from datetime import datetime, timezone

from qolsys.observable import QolsysObservable


LOGGER = logging.getLogger(__name__)


class QolsysPartition(QolsysObservable):

    NOTIFY_ADD_SENSOR = 'add_sensor'
    NOTIFY_REMOVE_SENSOR = 'remove_sensor'
    NOTIFY_UPDATE_ALARM_TYPE = 'update_alarm_type'
    NOTIFY_UPDATE_ATTRIBUTES = 'update_attributes'
    NOTIFY_UPDATE_SECURE_ARM = 'update_secure_arm'
    NOTIFY_UPDATE_STATUS = 'update_status'

    def __init__(self, partition_id: int, name: str, status: str,
                 secure_arm: bool) -> None:
        super().__init__()

        self._id = partition_id
        self._name = name
        self._status = status
        self._secure_arm = secure_arm
        self._sensors = {}
        self._alarm_type = None

        self._last_error_type = None
        self._last_error_desc = None
        self._last_error_at = None
        self._disarm_failed = 0

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def status(self):
        return self._status

    @property
    def secure_arm(self):
        return self._secure_arm

    @property
    def alarm_type(self):
        return self._alarm_type

    @property
    def sensors(self):
        return self._sensors.values()

    @property
    def last_error_type(self):
        return self._last_error_type

    @property
    def last_error_desc(self):
        return self._last_error_desc

    @property
    def last_error_at(self):
        return self._last_error_at

    @property
    def disarm_failed(self):
        return self._disarm_failed

    @status.setter
    def status(self, value):
        new_value = value.upper()
        if self._status != new_value:
            LOGGER.debug(f"Partition '{self.id}' ({self.name}) status updated to '{new_value}'")
            prev_value = self._status

            self._status = new_value

            self.notify(change=self.NOTIFY_UPDATE_STATUS,
                        prev_value=prev_value, new_value=new_value)

        # If the panel is disarmed, we can reset the failed disarm attempts
        if new_value.upper() == 'DISARM':
            self.disarm_failed = 0

        self.alarm_type = None

    @secure_arm.setter
    def secure_arm(self, value):
        new_value = bool(value)
        if self._secure_arm != new_value:
            LOGGER.debug(f"Partition '{self.id}' ({self.name}) secure arm updated to '{new_value}'")
            prev_value = self._secure_arm
            self._secure_arm = new_value

            self.notify(change=self.NOTIFY_UPDATE_SECURE_ARM,
                        prev_value=prev_value, new_value=new_value)

    @alarm_type.setter
    def alarm_type(self, value):
        if value is not None:
            value = value.upper()

        if self._alarm_type != value:
            LOGGER.debug(f"Partition '{self.id}' ({self.name}) alarm type updated to '{value}'")
            prev_value = self._alarm_type
            self._alarm_type = value

            self.notify(change=self.NOTIFY_UPDATE_ALARM_TYPE,
                        prev_value=prev_value, new_value=value)

    @disarm_failed.setter
    def disarm_failed(self, value):
        new_value = int(value)

        if self._disarm_failed != new_value:
            LOGGER.debug(f"Partition '{self.id}' ({self.name}) disarm failed updated to '{value}'")
            self._disarm_failed = new_value

            self.notify(change=self.NOTIFY_UPDATE_ATTRIBUTES)

    def triggered(self, alarm_type: str = None):
        self.status = 'ALARM'
        self.alarm_type = alarm_type

    def errored(self, error_type: str, error_description: str):
        self._last_error_type = error_type
        self._last_error_desc = error_description
        self._last_error_at = datetime.now(timezone.utc).isoformat()

        # If this is a failed disarm attempt, let's increase the counter
        if error_type.upper() == 'DISARM_FAILED':
            self._disarm_failed += 1

        self.notify(change=self.NOTIFY_UPDATE_ATTRIBUTES)

    def zone(self, zone_id, default=None):
        return self._sensors.get(zone_id, default)

    def sensor(self, sensor_id, default=None):
        for sensor in self._sensors.values():
            if sensor.id == sensor_id:
                return sensor

        return default

    def add_sensor(self, sensor):
        psensor = self._sensors.get(sensor.zone_id)
        if psensor is not None:
            LOGGER.error(f"Zone ID '{sensor.zone_id}' already used by sensor "
                         f"'{psensor.id}' ({psensor.name}) but sensor "
                         f"'{sensor.id}' ({sensor.name}) declares the same "
                         "zone; skipping")
            return

        self._sensors[sensor.zone_id] = sensor
        self.notify(change=self.NOTIFY_ADD_SENSOR, new_value=sensor)

    def update_sensor(self, sensor):
        psensor = self._sensors.get(sensor.zone_id)
        if psensor is None:
            return

        psensor.update(sensor)

    def remove_sensor(self, sensor):
        self.remove_zone(sensor.zone_id)

    def remove_zone(self, zone_id):
        zone = self._sensors[zone_id]

        del self._sensors[zone_id]

        self.notify(change=self.NOTIFY_REMOVE_SENSOR,
                    prev_value=zone)

    def __str__(self):
        return (f"<QolsysPartition id={self.id} name={self.name} "
                f"status={self.status} secure_arm={self.secure_arm} "
                f"sensors({len(self.sensors)})="
                f"[{', '.join([str(s) for s in self.sensors])}]>")



import logging

from mqtt.exceptions import MqttException

from qolsys.events import QolsysEventInfoSummary
from qolsys.exceptions import QolsysException
from qolsys.observable import QolsysObservable


LOGGER = logging.getLogger(__name__)


class QolsysState(QolsysObservable):
    NOTIFY_UPDATE_PARTITIONS = 'update_partitions'
    NOTIFY_UPDATE_ERROR = 'update_error'

    def __init__(self, event: QolsysEventInfoSummary = None):
        super().__init__()

        self._last_exception = None

        self._partitions = {}
        if event:
            self.update(event)

        QolsysException.STATE = self
        MqttException.STATE = self

    @property
    def last_exception(self):
        return self._last_exception

    @last_exception.setter
    def last_exception(self, value):
        prev_value = self._last_exception
        self._last_exception = value

        self.notify(change=self.NOTIFY_UPDATE_ERROR,
                    prev_value=prev_value,
                    new_value=value)

    @property
    def partitions(self):
        return self._partitions.values()

    def partition(self, partition_id):
        return self._partitions.get(int(partition_id))

    def update(self, event: QolsysEventInfoSummary):
        prev_partitions = self.partitions

        self._partitions = {}
        for partition in event.partitions:
            self._partitions[int(partition.id)] = partition

        self.notify(change=self.NOTIFY_UPDATE_PARTITIONS,
                    prev_value=prev_partitions,
                    new_value=self.partitions)

    def zone(self, zone_id):
        for partition in self.partitions:
            zone = partition.zone(zone_id)
            if zone is not None:
                return zone

    def sensor(self, sensor_id):
        for partition in self.partitions:
            sensor = partition.sensor(sensor_id)
            if sensor is not None:
                return sensor

    def zone_update(self, sensor):
        # Find where the zone is currently at
        current_zone = self.zone(sensor.zone_id)
        if current_zone is None:
            raise Exception(f'Zone not found for zone update: {sensor}, '
                            'we might not be sync-ed anymore')  # TODO: make it a better exception

        if current_zone.partition_id == sensor.partition_id:
            self._partitions[sensor.partition_id].update_sensor(sensor)
        else:
            self._partitions[current_zone.partition_id].remove_zone(sensor.zone_id)
            self._partitions[sensor.partition_id].add_sensor(sensor)

    def zone_add(self, sensor):
        partition = self._partitions.get(sensor.partition_id)
        if partition is None:
            raise Exception(f'Partition not found for zone add: {sensor},'
                            'we might not be sync-ed anymore')  # TODO: make it a better exception

        self._partitions[sensor.partition_id].add_sensor(sensor)

    def zone_open(self, zone_id):
        for partition in self.partitions:
            zone = partition.zone(zone_id)
            if zone is not None:
                zone.open()

    def zone_closed(self, zone_id):
        for partition in self.partitions:
            zone = partition.zone(zone_id)
            if zone is not None:
                zone.closed()



import json
import logging


LOGGER = logging.getLogger(__name__)


class QolsysAction(object):

    _PARAMS_TO_REDACT = []

    _DEFAULT_DATA = {
        'nonce': 'qolsys',
        'source': 'C4',
        'version': 0,
    }

    @property
    def data(self) -> dict:
        return {**self._DEFAULT_DATA, **self._data}

    @property
    def redacted(self) -> str:
        return json.dumps({
            **{k: '<redacted>' if k in self._PARAMS_TO_REDACT else v
               for k, v in self.data.items()},
            'token': '<redacted>'
        })

    def with_token(self, token) -> str:
        return json.dumps({**self.data, 'token': token})

    def __str__(self) -> str:
        return json.dumps(self.data)


class QolsysActionInfo(QolsysAction):
    def __init__(self) -> None:
        self._data = {
            'action': 'INFO',
            'info_type': 'SUMMARY',
        }


class QolsysActionArm(QolsysAction):

    ARMING_TYPE_DISARM = 'DISARM'
    ARMING_TYPE_ARM_AWAY = 'ARM_AWAY'
    ARMING_TYPE_ARM_STAY = 'ARM_STAY'

    _PARAMS_TO_REDACT = ['usercode']

    def __init__(self, arm_type: str, partition_id: int,
                 panel_code: str = None) -> None:
        self._data = {
            'action': 'ARMING',
            'arming_type': arm_type,
            'partition_id': partition_id,
        }

        if panel_code:
            self._data['usercode'] = str(panel_code)


class QolsysActionArmWithDelayAndBypass(QolsysActionArm):
    def __init__(self, delay: int = None, bypass: bool = None,
                 *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if delay is not None and delay >= 0:
            self._data['delay'] = delay

        if bypass is not None:
            self._data['bypass'] = str(bypass).lower()


class QolsysActionDisarm(QolsysActionArm):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(arm_type=QolsysActionArm.ARMING_TYPE_DISARM,
                         *args, **kwargs)


class QolsysActionArmAway(QolsysActionArmWithDelayAndBypass):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(arm_type=QolsysActionArm.ARMING_TYPE_ARM_AWAY,
                         *args, **kwargs)


class QolsysActionArmStay(QolsysActionArmWithDelayAndBypass):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(arm_type=QolsysActionArm.ARMING_TYPE_ARM_STAY,
                         *args, **kwargs)


class QolsysActionTrigger(QolsysAction):
    ALARM_TYPE_POLICE = 'POLICE'
    ALARM_TYPE_FIRE = 'FIRE'
    ALARM_TYPE_AUXILIARY = 'AUXILIARY'

    def __init__(self, partition_id: int, alarm_type: str = None) -> None:
        self._data = {
            'action': 'ALARM',
            'alarm_type': alarm_type or self.ALARM_TYPE_POLICE,
            'partition_id': partition_id,
        }



import json
import logging

from qolsys.actions import QolsysActionArmAway
from qolsys.actions import QolsysActionArmStay
from qolsys.actions import QolsysActionDisarm
from qolsys.actions import QolsysActionTrigger
from qolsys.exceptions import UnknownQolsysControlException
from qolsys.exceptions import MissingUserCodeException
from qolsys.exceptions import InvalidUserCodeException
from qolsys.utils import find_subclass


LOGGER = logging.getLogger(__name__)


class QolsysControl(object):

    __SUBCLASSES_CACHE = {}

    def __init__(self, raw: dict, partition_id: int, code: str = None,
                 session_token: str = None):
        self._raw = raw
        self._partition_id = partition_id
        self._session_token = session_token

        self._code = code
        if self._code is not None:
            self._code = str(self._code)

        self._requires_config = False

    @property
    def partition_id(self):
        return self._partition_id

    @property
    def code(self):
        return self._code

    @property
    def session_token(self):
        return self._session_token

    @property
    def raw(self):
        return self._raw

    @property
    def raw_str(self):
        return json.dumps(self.raw)

    @property
    def requires_config(self):
        return self._requires_config

    def configure(self, cfg, state):
        pass

    def check(self):
        pass

    @property
    def action(self):
        return None

    def __str__(self):
        return (f"<{type(self).__name__} partition_id={self.partition_id} "
                f"code={'<redacted>' if self._code else self._code} "
                f"session_token={self.session_token}>")

    @classmethod
    def from_json(cls, data):
        if isinstance(data, str):
            data = json.loads(data)

        action_type = data.get('action')
        klass = find_subclass(QolsysControl, action_type,
                              cache=QolsysControl.__SUBCLASSES_CACHE)
        if not klass:
            raise UnknownQolsysControlException(
                "Unable to find a QolsysControl class for "
                f"action '{action_type}'")

        from_json = getattr(klass, 'from_json', None)
        if callable(from_json) and \
                from_json.__func__ != QolsysControl.from_json.__func__:
            return from_json(data)

        if hasattr(klass, '_EXTRA_PARAMS'):
            extra_params = {
                p: data.get(p)
                for p in klass._EXTRA_PARAMS
            }
        else:
            extra_params = {}

        return klass(
            partition_id=data.get('partition_id'),
            code=data.get('code'),
            session_token=data.get('session_token'),
            raw=data,
            **extra_params,
        )


class _QolsysControlCheckCode(QolsysControl):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._requires_config = True
        self._panel_code = None

    def configure(self, cfg, state):
        self._secure_arm = hasattr(self, '_partition_id') and \
            state.partition(self._partition_id).secure_arm

        if self._PANEL_CODE_REQUIRED == 'secure_arm':
            self._PANEL_CODE_REQUIRED = self._secure_arm

        self._panel_code = cfg.panel_user_code

        code_required_from_ha = getattr(cfg, self._CODE_REQUIRED_ATTR)
        self._check_code = code_required_from_ha and not cfg.ha_check_user_code
        if self._check_code:
            self._valid_code = cfg.ha_user_code or cfg.panel_user_code

    def check(self):
        super().check()

        if self._check_code:
            # The only condition where _valid_code is None and we still
            # require to get a code from home assistant is if we are in
            # the disarm process; in which case, we don't want to raise
            # an exception, as we want to try and use that provided code
            # to disarm the alarm
            if self._valid_code and self._code != self._valid_code:
                raise InvalidUserCodeException(
                    'Code received in the control command invalid')

        if self._PANEL_CODE_REQUIRED:
            LOGGER.debug(f"Panel code is required for {self}")
            if self._panel_code is None:
                LOGGER.debug('Panel code is not provided in config')
                if self._code:
                    LOGGER.info('Using code sent from home assistant since '
                                'no panel code configured')
                    self._panel_code = self._code
                else:
                    raise MissingUserCodeException(
                        'Cannot perform action without a configured panel code')


class QolsysControlDisarm(_QolsysControlCheckCode):

    _CODE_REQUIRED_ATTR = 'code_disarm_required'
    _PANEL_CODE_REQUIRED = True

    @property
    def action(self):
        return QolsysActionDisarm(
            partition_id=self._partition_id,
            panel_code=self._panel_code,
        )


class QolsysControlArm(_QolsysControlCheckCode):
    _CODE_REQUIRED_ATTR = 'code_arm_required'
    _PANEL_CODE_REQUIRED = 'secure_arm'
    _EXTRA_PARAMS = [
        'bypass',
        'delay',
    ]

    def __init__(self, delay: int = None, bypass: bool = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._delay = delay
        self._bypass = bypass

        self._requires_config = (self._requires_config or
                                 delay is None or
                                 bypass is None)

    def configure(self, cfg, state):
        super().configure(cfg, state)

        if self._delay is None:
            self._delay = getattr(cfg, f'{self._ATTR_PREFIX}_exit_delay')

        if self._bypass is None:
            self._bypass = getattr(cfg, f'{self._ATTR_PREFIX}_bypass')

    @property
    def action(self):
        return self._ACTION_CLASS(
            partition_id=self._partition_id,
            panel_code=self._panel_code,
            delay=self._delay,
            bypass=self._bypass,
        )


class QolsysControlArmAway(QolsysControlArm):
    _ATTR_PREFIX = 'arm_away'
    _ACTION_CLASS = QolsysActionArmAway


class QolsysControlArmVacation(QolsysControlArmAway):
    pass


class QolsysControlArmHome(QolsysControlArm):
    _ATTR_PREFIX = 'arm_stay'
    _ACTION_CLASS = QolsysActionArmStay


class QolsysControlArmNight(QolsysControlArmHome):
    pass


class QolsysControlArmCustomBypass(QolsysControlArm):
    _ACTION_CLASSES = {
        'arm_stay': QolsysActionArmStay,
        'arm_away': QolsysActionArmAway,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(bypass=True, *args,
                         **{k: v for k, v in kwargs.items() if k != 'bypass'})

        self._require_config = True

    def configure(self, cfg, state):
        self._ATTR_PREFIX = cfg.arm_type_custom_bypass.lower()
        self._ACTION_CLASS = self._ACTION_CLASSES[self._ATTR_PREFIX]

        super().configure(cfg, state)


# This depends on https://github.com/home-assistant/core/pull/60525, which
# has been merged and is available starting with Home Assitant 2021.12
class QolsysControlTrigger(_QolsysControlCheckCode):

    _CODE_REQUIRED_ATTR = 'code_trigger_required'
    _PANEL_CODE_REQUIRED = False

    def __init__(self, alarm_type: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._alarm_type = alarm_type

    @property
    def action(self):
        return QolsysActionTrigger(
            partition_id=self._partition_id,
            alarm_type=self._alarm_type,
        )


class QolsysControlTriggerPolice(QolsysControlTrigger):
    def __init__(self, *args, **kwargs):
        super().__init__(alarm_type=QolsysActionTrigger.ALARM_TYPE_POLICE,
                         *args, **kwargs)


class QolsysControlTriggerFire(QolsysControlTrigger):
    def __init__(self, *args, **kwargs):
        super().__init__(alarm_type=QolsysActionTrigger.ALARM_TYPE_FIRE,
                         *args, **kwargs)


class QolsysControlTriggerAuxiliary(QolsysControlTrigger):
    def __init__(self, *args, **kwargs):
        super().__init__(alarm_type=QolsysActionTrigger.ALARM_TYPE_AUXILIARY,
                         *args, **kwargs)



import logging
import re
import subprocess


LOGGER = logging.getLogger(__name__)


class LoggerCallback(object):
    def __init__(self, msg=None):
        self.msg = msg or 'Logger callback'

    async def __call__(self, *args, **kwargs):
        LOGGER.debug(f"{self.msg}{': ' if args or kwargs else ''}"
                     f"{f'args={args} ' if args else ''}"
                     f"{f'kwargs={kwargs}' if kwargs else ''}")


async def defaultLoggerCallback(*args, **kwargs):
    callback = LoggerCallback()
    callback(*args, **kwargs)


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def find_subclass(cls, subtype: str, cache: dict = None, normalize=True,
                  preserve_capitals=False):
    if cache and subtype in cache:
        return cache[subtype]

    normalized_subtype = subtype
    if normalize:
        normalized_subtype = re.compile(r'[\W_]+').sub(' ', normalized_subtype)
        if preserve_capitals:
            normalized_subtype = re.compile(r'(?<=[^\s])([A-Z])').sub(
                ' \\1', normalized_subtype)
        normalized_subtype = normalized_subtype.title()
        normalized_subtype = re.compile(r'\s').sub('', normalized_subtype)

    search = f"{cls.__name__}{normalized_subtype}"

    for klass in all_subclasses(cls):
        if klass.__name__ == search:
            if cache is not None:
                cache[subtype] = klass
            return klass

    if cache is not None:
        cache[subtype] = None
    return None


def get_mac_from_host(ip_or_host):
    try:
        # The arp command will automatically resolve the hostname to an
        # IP address for us if needed
        process = subprocess.run(['arp', ip_or_host], capture_output=True)
    except subprocess.SubprocessError:
        LOGGER.exception(f"Error trying to get the mac address for '{ip_or_host}'")
        return None

    output = process.stdout.decode('utf-8')

    m = re.search(r'(([a-f\d]{1,2}:){5}[a-f\d]{1,2})', output)

    if m is None:
        LOGGER.warning(f"No mac address found for '{ip_or_host}'")
        return None

    mac = m[0]
    LOGGER.debug(f"Found mac address '{mac}' for '{ip_or_host}'")
    return mac



from datetime import datetime, timezone


class QolsysException(Exception):
    STATE = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._at = datetime.now(timezone.utc).isoformat()

        if self.STATE:
            self.STATE.last_exception = self

    @property
    def at(self):
        return self._at


class QolsysGwConfigIncomplete(QolsysException):
    pass


class QolsysGwConfigError(QolsysException):
    pass


class UnableToParseEventException(QolsysException):
    pass


class UnableToParseSensorException(QolsysException):
    pass


class UnknownQolsysControlException(QolsysException):
    pass


class UnknownQolsysEventException(QolsysException):
    pass


class UnknownQolsysSensorException(QolsysException):
    pass


class MissingUserCodeException(QolsysException):
    pass


class InvalidUserCodeException(QolsysException):
    pass
